package com.igzcode.java.web.data;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.commons.dbcp.BasicDataSource;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.igzcode.java.util.data.DbType;
import com.igzcode.java.web.util.ConfigUtil;

/**
 * Database Layer, this work with SQL ResultSets and return DTOs.
 * It is used by AbstractFactory class.
 * @param <DtoType> A POJO with field names equals to the database field names
 */
public class Sql<DtoType> {
	
	static private final Logger _LOG = Logger.getLogger("com.igzcode.java.web.data.Sql");
	
	static private HashMap<String, BasicDataSource> _DataSources;
	
	private Connection _Conn;
	private String _ConnName;
	
	static private HashMap<String, HashMap<String, QueryInfo>> _QueryMap = null; // FactoryName -> QueryId -> QueryInfo
	
	static {
		_LOG.info("Loading factory XML files...");
		
		File sqlFolder = new File(Sql.class.getClassLoader().getResource("sql").getFile());
		if ( sqlFolder.isDirectory() ) {
			File[] sqlFiles = sqlFolder.listFiles();
			_QueryMap = new HashMap<String, HashMap<String, QueryInfo>>();
			for( File sqlFile : sqlFiles ) {
				try {
					_LoadQueryFile( sqlFile );
					_LOG.info(" - " + sqlFile.getName() + " loaded." );
				}
				catch (Exception e) {
					e.printStackTrace();
					_LOG.severe(" - " + sqlFile.getName() + " error loading file.");
				}
			}
		}
	}

	public Sql ( String p_connName ) {
		_ConnName = p_connName;
		
		if ( _DataSources == null ) {
			_DataSources = new HashMap<String, BasicDataSource>();
		}
		
		if ( ! _DataSources.containsKey(p_connName) ) {
			
			BasicDataSource bds = new BasicDataSource();
			bds.setUsername( ConfigUtil.Env(p_connName, "usr") );
			bds.setPassword( ConfigUtil.Env(p_connName, "pwd") );
			
			String server = ConfigUtil.Env(p_connName, "server");
			String dbname = ConfigUtil.Env(p_connName, "dbname");
			
			if ( ConfigUtil.IsTestModeEnabled() ) {
				dbname += "_test";
			}
			
			switch ( DbType.valueOf( ConfigUtil.Env(p_connName, "dbtype") ) ) {
				case MYSQL:
					bds.setDriverClassName("com.mysql.jdbc.Driver");
					bds.setUrl( "jdbc:mysql://" + server + "/" + dbname );
					bds.setValidationQuery("select 1"); // To avoid: SQL Exception - Connection is already closed, for more details visit: http://community.jboss.org/message/88645
					break;
					
				case SQLSERVER:
					bds.setDriverClassName("com.microsoft.sqlserver.jdbc.SQLServerDriver");
					bds.setUrl( "jdbc:sqlserver://" + server + ";DatabaseName=" + dbname );
					break;
			}
			
			_DataSources.put(p_connName, bds);
		}
		
	}
	
	private void _ConnInit () throws SQLException {
		if ( _Conn == null ) {
			_Conn = _DataSources.get(_ConnName).getConnection();
		}
	}
	
	private void _ConnClose (Statement p_ps) throws SQLException {
		if ( p_ps!=null && !p_ps.isClosed() ) {
			p_ps.close();
		}
		
		if ( _Conn != null && !_Conn.isClosed() ) {
			_Conn.close();
			_Conn = null;
		}
	}
	
	public Integer Insert ( String p_sqlInsert, Object[] p_values ) throws SQLException {
		_ConnInit();

		Integer generatedKey = null;
		
		PreparedStatement ps = _Conn.prepareStatement(p_sqlInsert, Statement.RETURN_GENERATED_KEYS);
		try {
			_SetValuesToPS( p_values, ps );
			ps.executeUpdate();
			
			ResultSet rsKeyColumn = ps.getGeneratedKeys();
			if ( rsKeyColumn.next() ) {
				generatedKey = new Integer( rsKeyColumn.getInt(1) );
	        }
		}
		catch ( SQLException e ) {
			e.printStackTrace();
			throw e;
		}
		finally {
			_ConnClose(ps);
		}
		
		return generatedKey;		
	}
	
	public Integer Update ( String p_sqlUpdate, Object[] p_values ) throws SQLException {
		_ConnInit();
		
		Integer numRows = null;
		
		PreparedStatement ps = _Conn.prepareStatement( p_sqlUpdate );
    	try {
    		_SetValuesToPS(p_values, ps);
    		numRows = ps.executeUpdate();
    	}
    	catch ( SQLException e ) {
    		throw e;
    	}
    	finally {
    		_ConnClose(ps);
    	}
	    	
		return numRows;
	}
	
	public int Delete ( String p_sql ) throws SQLException {
		_ConnInit();
		
		int numRows = 0;
		
		Statement statement = _Conn.createStatement();
		try {
			numRows = statement.executeUpdate( p_sql );
		}
		catch ( SQLException e ) {
			throw e;
		}
		finally {
			_ConnClose(statement);
		}
		
		return numRows;
	}
	
	public List<DtoType> Query (String p_factoryName, String p_queryId, Map<String, Object> p_params, Class<DtoType> p_class) throws SQLException {

		QueryInfo queryInfo = _QueryMap.get(p_factoryName).get(p_queryId);
		String sql = queryInfo.GetSql();
		
		List<String> paramOrder = queryInfo.GetParamOrder();
		if ( paramOrder.size() > 0 ) {
			
			// Parameterized query
			Object[] params = new Object[paramOrder.size()];
			for ( int i = 0; i < paramOrder.size(); i++ ) {
				params[i] = p_params.get( paramOrder.get(i) );
			}
			return ExecuteQueryString(sql, params, p_class);
			
		}
		else {
			return ExecuteQueryString(sql, null, p_class);
		}
	}

	public ArrayList<DtoType> ExecuteQueryString (String p_parameterizedQuery, Object[] p_paramValues, Class<DtoType> p_class ) throws SQLException {
		_ConnInit();
		
		ArrayList<DtoType> list = new ArrayList<DtoType>();
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			ps = _Conn.prepareStatement( p_parameterizedQuery );
		
			_SetValuesToPS(p_paramValues, ps);
			rs = ps.executeQuery();

			DtoType dto;
			Field[] fields = p_class.getDeclaredFields();
			ArrayList<Field> validFields = new ArrayList<Field>();
			int f, F = fields.length;
			for( f=0; f<F; f++ ) {
				fields[f].setAccessible(true);
				try {
					rs.findColumn( fields[f].getName().substring(1) ); // Remove "_" prefix
				}
				catch (Exception e) {
					continue; // If a DTO field not exists in the ResultSet
				}
				validFields.add( fields[f] );
			}
			
			F = validFields.size();
			String fieldName;
			Object value;
			GregorianCalendar calendar;
			Class<?> fieldType;
			while( rs.next() ) {
				dto = p_class.newInstance();
				for( f=0; f<F; f++ ) {
					fieldType = validFields.get(f).getType();
					fieldName = validFields.get(f).getName().substring(1); // Remove "_" prefix
					
					value = rs.getObject( fieldName );
					
					if ( value != null ){
						
						if ( fieldType == Integer.class ) {						
							validFields.get(f).set( dto, rs.getInt( fieldName ) );
						}
						else if ( fieldType == Calendar.class ) {
							calendar = new GregorianCalendar();
							calendar.setTimeInMillis( ((Timestamp) value).getTime() );
							validFields.get(f).set( dto, calendar );
						}
						else if ( fieldType == Float.class ) {						
							validFields.get(f).set( dto, rs.getFloat( fieldName ) );
						}
						else if ( fieldType == Double.class ) {						
							validFields.get(f).set( dto, rs.getDouble( fieldName ) );
						}
						else if ( fieldType == Long.class ) {						
							validFields.get(f).set( dto, rs.getLong( fieldName ) );
						}
						else {
							validFields.get(f).set( dto, value );
						}
					}
				}
				list.add( dto );
			}
		}
		catch (IllegalArgumentException e) {
			e.printStackTrace();
		}
		catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		catch (InstantiationException e) {
			e.printStackTrace();
		}
		
		finally {
			if ( rs!=null && !rs.isClosed()) {
				rs.close();
			}
			_ConnClose(ps);
		}

		return list;
	}
	
	private void _SetValuesToPS ( Object[] p_values, PreparedStatement p_ps ) throws SQLException {
		if ( p_values != null && p_values.length > 0 ) {
			Object value;
			int f, F = p_values.length;
			for( f=0; f<F; f++ ) {
				value = p_values[f];
				if ( value != null && value.getClass() == GregorianCalendar.class ) {
					p_ps.setTimestamp( f+1, new Timestamp( ((Calendar) value).getTime().getTime() ) );
	            }
				else {
	            	p_ps.setObject( f+1, value );
	            }
			}
		}
	}
	
	static private void _LoadQueryFile ( File p_file ) throws ParserConfigurationException, SAXException, IOException, XPathExpressionException {
		
		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
		Document doc = docBuilder.parse( p_file );

		// Prepare XPath
		XPathFactory factory = XPathFactory.newInstance();
		XPath xpath = factory.newXPath();
		
		// Load data
		XPathExpression expression = xpath.compile("/Queries/*");
		NodeList nodeList = (NodeList) expression.evaluate(doc, XPathConstants.NODESET);

		HashMap<String, QueryInfo> queryInfos = new HashMap<String, QueryInfo>();
		_QueryMap.put(p_file.getName().replace(".xml", ""), queryInfos);
		
		Node node;
		String sql;
		Matcher matcher;
		ArrayList<String> paramOrder;
		Pattern pattern = Pattern.compile("((=|\\+){1}\\s*)@([a-zA-Z]+)");
		for ( int i=0; i<nodeList.getLength(); i++ ) {
			node = nodeList.item(i);
			sql = node.getTextContent().trim().replaceAll("[\\t\\n\\r]", " ");
			
			// Get query parameters
			paramOrder = new ArrayList<String>();
			matcher = pattern.matcher(sql);
			while ( matcher.find() ) {
				paramOrder.add( matcher.group(3) );
			}
			sql = sql.replaceAll(pattern.toString(), "$1?");
			
			queryInfos.put(node.getNodeName(), new QueryInfo(node.getNodeName(), sql, paramOrder));
		}
		
	}
}
