package com.igzcode.java.web.pattern;

import java.lang.reflect.Field;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import com.igzcode.java.web.data.Sql;
import com.igzcode.java.web.data.Transient;

/**
 * Abstract class to implement the DTO-Factory-Manager pattern.
 * 
 * @param <DtoType> A POJO with field names equals to the database field names
 */
public abstract class AbstractFactory <DtoType> {

	static private final Logger _LOG = Logger.getLogger("com.igzcode.java.web.pattern.AbstractFactory");


	protected String _FactoryClassName 	= null;
	protected String _ConnName 		= null;
	protected String _TableName 	= null;
	protected String _PrimaryKey 	= null;
	protected Sql<DtoType> _Sql = null;

	protected Class<DtoType> _DtoClassType = null;
	private Field[] _DtoFields = null;
	private Field _PrimaryKeyField = null;
	

	/**
	 * Set the needed information to data access and deserialize DTOs.
	 * 
	 * @param p_connName Connection name
	 * @param p_tableName Table name
	 * @param p_primaryKey Primary key name
	 * @param p_class DTO Class reference
	 */
	protected AbstractFactory ( String p_connName, String p_tableName, String p_primaryKey, Class<DtoType> p_class ) {
		_ConnName = p_connName;
		_TableName = p_tableName;
		_Sql	= new Sql<DtoType>(_ConnName);
		_PrimaryKey = p_primaryKey;
		_DtoClassType = p_class;
		_DtoFields = _DtoClassType.getDeclaredFields();
		_FactoryClassName = getClass().getSimpleName().replace("Manager", "Factory");
		
		try {
			_PrimaryKeyField = _DtoClassType.getDeclaredField("_" + _PrimaryKey);
			_PrimaryKeyField.setAccessible(true);
		}
		catch (Exception e) {
			e.printStackTrace();
			_LOG.severe("The primary key field [_" + _PrimaryKey + "] not exists into DtoType");
		}
	}
	
	
	/**
	 * Execute a query without parameters.<br>
	 * These queries are defined into files named {@code ClassName.xml }
	 * at {@code resources/sql} folder.<br>
	 * A quick example, for the class {@code ExampleFactory.java} you must
	 * create {@code resources/sql/ExampleFactory.xml}:
	 * <pre>
	 * {@code
	 * <?xml version="1.0" encoding="utf-8"?>
	 * <Queries>
	 * 	  <FindAll>
	 * 	   <![CDATA[
	 * 		   SELECT *
	 * 		   FROM table_name
	 * 	   ]]>
	 * 	  </FindAll>
	 * </Queries>
	 * }
	 * </pre>
	 * 
	 * To use the before query:
	 * <pre>
	 * {@code _Query("FindAll"); }
	 * </pre>
	 * 
	 * @param p_id Query identifier (node name of XML file)
	 * @return A List with the DTOs result
	 * @throws SQLException
	 */
	protected List<DtoType> _Query ( String p_id ) throws SQLException {
		return _Query( p_id, null );
	}
	
	/**
	 * Execute a parameterized query.<br>
	 * These queries are defined into files like {@code ClassName.xml}
	 * at {@code resources/sql} folder.<br>
	 * A quick example, for the class {@code ExampleFactory.java} you must
	 * create {@code resources/sql/ExampleFactory.xml}:
	 * <pre>
	 * {@code
	 * <?xml version="1.0" encoding="utf-8"?>
	 * <Queries>
	 * 	  <FindAll>
	 * 	   <![CDATA[
	 * 		   SELECT *
	 * 		   FROM table_name
	 * 		   WHERE id = @id
	 * 	   ]]>
	 * 	  </FindAll>
	 * </Queries>
	 * }
	 * </pre>
	 * 
	 * To use the before query:
	 * <pre>
	 * {@code
	 * HashMap<String, Object> params = new HashMap<String, Object>();
	 * params.put("id", 1022);
	 * _Query("FindAll", params);
	 * }
	 * </pre>
	 * 
	 * @param p_id Node name of xml file
	 * @param p_params Map with parameter values classified by parameter name
	 * @return A List with the serialized DTOs result
	 * @throws SQLException
	 */
	protected List<DtoType> _Query (String p_id, Map<String, Object> p_params ) throws SQLException {
		return _Sql.Query(_FactoryClassName, p_id, p_params, _DtoClassType);
	}
	
	/**
	 * Execute a SQL query string.
	 * <pre>
	 * _Query( "SELECT * FROM table_name"
	 *    + "WHERE name = " 
	 *    + StringUtil.ParseToSQL(name) 
	 * );
	 * </pre>
	 * @see {@link  com.igzcode.java.util.StringUtil#ParseToSQL  ParseToSQL}
	 * @see {@link  com.igzcode.java.util.StringUtil#ParseToSQL  ParseToSQLLike}
	 * @param p_queryString Query string
	 * @return A List with the serialized DTOs result
	 * @throws SQLException
	 */
	protected List<DtoType> _CustomQuery (String p_queryString ) throws SQLException {
		return _CustomQuery(p_queryString, null);
	}
	
	/**
	 * Execute a parameterized query string.<br>
	 * Parameters must be ordered by query apparition order.
	 * <pre>
	 * {@code
	 * Object[] params = new Object[2];
	 * params[0] = "John";
	 * params[1] = 127;
	 * 
	 * _Query( "SELECT * FROM table_name"
	 *     + "WHERE name = ? OR id = ?"
	 *     , params
	 * );
	 * }
	 * </pre>
	 * 
	 * @param p_parameterizedQuery Parameterized query string
	 * @param p_params Object array with the ordered parameter values
	 * @return A List with the serialized DTOs result
	 * @throws SQLException
	 */
	protected List<DtoType> _CustomQuery (String p_parameterizedQuery, Object[] p_params ) throws SQLException {
		return _Sql.ExecuteQueryString(p_parameterizedQuery, p_params, _DtoClassType);
	}
	
	/**
	 * Find a DTO by identifier.
	 * If the DTO do not exist, return null.
	 * <pre>
	 * {@code
	 * DtoType dto = _Get(127);
	 * }
	 * </pre>
	 * @param p_id DTO identifier
	 * @return A saved DTO or {@code null} if not exists
	 * @throws SQLException
	 */
	protected DtoType _Get ( int p_id ) throws SQLException {
		DtoType dto = null;
		
		List<DtoType> queryResult = _CustomQuery("SELECT * FROM " + _TableName + " WHERE " + _PrimaryKey + "=" + p_id);
		if ( queryResult != null && queryResult.size() == 1 ) {
			dto = queryResult.get(0);
		}
		
		return dto;
	}
	
	
	/**
	 * Get multiple of DTO.
	 * <pre>
	 * {@code
	 * 
	 * List<Integer> ids = new ArrayList<Integer>();
	 * ids.add(127);
	 * ids.add(544);
	 * 
	 * List<DtoType> dtos = _Get(ids);
	 * }
	 * </pre>
	 * @param p_ids List of DTO identifiers
	 * @return A list of found DTOs
	 * @throws SQLException
	 */
	protected List<DtoType> _Get ( List<Integer> p_ids ) throws SQLException {
		return _CustomQuery(
			  "SELECT * FROM " 
			+ _TableName 
			+ " WHERE "
			+ _PrimaryKey
			+ " IN " + p_ids.toString().replace('[', '(').replace(']', ')')
		);
	}
	
	/**
	 * Retrieves all saved DTOs.
	 * @return All saved DTOs
	 * @throws SQLException
	 */
	protected List<DtoType> _FindAll () throws SQLException {
		return _CustomQuery("SELECT * FROM " + _TableName);
	}
	
	
	
	
	
	/**
	 * Delete a DTO by identifier.
	 * <pre>
	 * if ( _Delete(127) == 1 ) {
	 *   System.out.println("DTO deleted");
	 * }
	 * </pre>
	 * @param p_id DTO identifier
	 * @return The number of rows deleted
	 * @throws SQLException
	 */
	protected int _Delete ( int p_id ) throws SQLException {
        return _Delete( _PrimaryKey + "=" + p_id );
	}
	
	/**
	 * Delete multiple DTOs by identifiers in a single execution.
	 * <pre>
	 * List<Integer> ids = new ArrayList<Integer>();
	 * ids.add(127);
	 * ids.add(201);
	 * 
	 * int num_del = _Delete(ids);
	 * 
	 * System.out.println(num_del + " DTOs deleted");
	 * </pre>
	 * @param p_ids DTO identifiers
	 * @return The number of rows deleted
	 * @throws SQLException
	 */
	protected int _Delete ( List<Integer> p_ids ) throws SQLException {
        return _Delete( _PrimaryKey + " IN (" + p_ids.toString().replace("[", "").replace("]", "") + ")" );
	}
	
	/**
	 * Delete one or multiple DTOs by a WHERE clause.
	 * <pre>
	 * String where = "FkUserId=34 AND FkDepartamentId=467";
	 * int num_del = _Delete(where);
	 * System.out.println(num_del + " DTOs deleted");
	 * </pre>
	 * @param p_whereClause
	 * @return The number of rows deleted
	 * @throws SQLException
	 */
	protected int _Delete ( String p_whereClause ) throws SQLException {
		return _Sql.Delete( "DELETE FROM " + _TableName + " WHERE " + p_whereClause );
	}
	
	/**
	 * <b>Warning:</b> Delete all entities.
	 * @return The number of rows deleted
	 * @throws SQLException
	 */
	protected int _DeleteAll () throws SQLException {
		return _Sql.Delete( "DELETE FROM " + _TableName );
	}
	
	
	/**
	 * Create or update a DTO.<br>
	 * If the primary key field has a value, you are executing a SQL UPDATE,
	 * otherwise you are executing a SQL CREATE.<br><br>
	 * <b>Important!</b> If you are updating, the saved values are not updated with null values,
	 * to enable this behaviour use {@code _Save(dto, true)} or  {@code _Update(dto, true)}.
	 * <br>
	 * <pre>
	 * int id = _Save(dtoInstance);
	 * </pre>
	 * @param p_dto DTO instance to save
	 * @return The DTO identifier
	 * @throws IllegalAccessException
	 * @throws SQLException
	 */
	protected int _Save ( DtoType p_dto ) throws SQLException, IllegalAccessException {
		return _Save( p_dto, false );
	}
	
	/**
	 * Create or update a DTO.
	 * If the primary key field has a value, you are executing a SQL UPDATE command,
	 * otherwise you are executing a SQL CREATE command.
	 * If you are updating, the saved values are updated with nulls if p_updateFillNulls is true.
	 * <pre>
	 * {@code
	 * // Update with null values 
	 * int id = _Save(dtoInstance, true);
	 * }
	 * </pre>
	 * @param p_dto DTO values to save
	 * @param p_updateFillNulls
	 * @return The DTO identifier
	 * @throws SQLException
	 * @throws IllegalAccessException
	 */
	protected int _Save ( DtoType p_dto, boolean p_updateFillNulls ) throws SQLException, IllegalAccessException {
		int id;
		
		if ( _PrimaryKeyField.get(p_dto) == null ) {
			id = _Create( p_dto );
		}
		else {
			id = _Update( p_dto, p_updateFillNulls );
		}
		
		return id;
	}

	/**
	 * Make persistent a new DTO values.
	 * This will execute a SQL INSERT INTO command.
	 * <pre>
	 * int id = _Create(dtoInstance);
	 * </pre>
	 * @param p_dto DTO instance
	 * @return A DTO identifier
	 * @throws SQLException
	 * @throws IllegalAccessException
	 */
	protected Integer _Create ( DtoType p_dto ) throws SQLException, IllegalAccessException {
		StringBuilder columnNames 	= new StringBuilder();
        StringBuilder columnValues 	= new StringBuilder();
        
        String key;
        Object value;
        ArrayList<Object> values = new ArrayList<Object>();
        
		for( int f=0, F = _DtoFields.length; f<F; f++ ) {
			_DtoFields[f].setAccessible(true);
					
			if ( !_DtoFields[f].isAnnotationPresent(Transient.class) ) {
				
				key = (String) _DtoFields[f].getName().substring(1);
				value = _DtoFields[f].get( p_dto );
				
	            if ( value != null ) {
	                columnNames.append(key + ",");
	                columnValues.append("?,");
	                values.add(value);
	            }
			}
		}
		
		columnNames.deleteCharAt(columnNames.length() - 1);
		columnValues.deleteCharAt(columnValues.length() - 1);
		
		StringBuilder insert = new StringBuilder("INSERT INTO ");
		insert.append(_TableName);
		insert.append(" (");
		insert.append(columnNames.toString());
		insert.append(") VALUES (");
		insert.append(columnValues.toString());
		insert.append(")");
		
		String sql = insert.toString();

		_LOG.info(sql + values.toString());
		
		return _Sql.Insert( sql, values.toArray() );
	}


	/**
	 * Update values of a created DTO.
	 * The DTO must have a identifier.
	 * The field names that start with "Ns" (Not Save) prefix are not saved.
	 * If you are updating, the saved values are updated with nulls only when p_updateFillNulls is true.
	 * <pre>
	 * int num_rows = _Update(dtoInstance, true);
	 * </pre>
	 * @param p_dto DTO to be updated
	 * @param p_fillNulls Indicates if perform a update with null values
	 * @return The number of rows updated
	 * @throws SQLException
	 * @throws IllegalAccessException
	 */
	protected Integer _Update ( DtoType p_dto, boolean p_fillNulls ) throws SQLException, IllegalAccessException {
		ArrayList<Object> values = new ArrayList<Object>();
		StringBuffer update = new StringBuffer("UPDATE " + _TableName + " SET ");

        Object keyValue = null;
        String fieldName;
        Object fieldValue;
        int f, F = _DtoFields.length;
		for( f=0; f<F; f++ ) {
			_DtoFields[f].setAccessible(true);
			
			fieldName  = _DtoFields[f].getName().substring(1);
            fieldValue = _DtoFields[f].get( p_dto );
			
            if( !_DtoFields[f].isAnnotationPresent(Transient.class) && (p_fillNulls || fieldValue != null) ) {
            	if ( !_PrimaryKey.equals(fieldName) ){
            		update.append(fieldName + "=?,");
	            	values.add( fieldValue );
            	}
            	else {
            		keyValue = fieldValue;
            	}
            }
		}
		update.deleteCharAt( update.length()-1 );

		if ( values.size() == 0 ) {
			_LOG.warning("SQL Update can not be executed because there are not parameters to update");
			return 0;
		}
		else {
			update.append(" WHERE "+ _PrimaryKey + "=?");
			values.add( keyValue );
			
			String query = update.toString();
			
			_LOG.info(query + values.toString());
			
			return _Update( query, values );
		}
	}
	
	/**
	 * Execute a SQL update string.
	 * A quick simple example:
	 * <pre>
	 * _Update( "UPDATE " + _TableName 
	 *        + " SET Deleted=1 WHERE " 
	 *        + _PrimaryKey 
	 *        + " IN (5,6)"
	 *     ,null
	 * );
	 * </pre>
	 * 
	 * Parameterized query example:
	 * <pre>
	 * ArrayList<Object> params = new ArrayList<Object>();
	 * params.add("New text");
	 * params.add(127);
	 * 
	 * _Update( "UPDATE " + _TableName + " SET Text=? WHERE " + _PrimaryKey + " = ?", params );
	 * </pre>
	 * 
	 * @param p_sqlUpdate UPDATE SQL string
	 * @param p_whereParams SQL Where parameters
	 * @return The number of rows updated
	 * @throws SQLException
	 */
	protected Integer _Update ( String p_sqlUpdate, List<Object> p_whereParams ) throws SQLException {
		
		_LOG.info(p_sqlUpdate + p_whereParams.toString());
		
		return _Sql.Update( p_sqlUpdate, p_whereParams.toArray() );
	}
	
}