package com.igzcode.java.web.util;

import java.io.IOException;
import java.util.HashMap;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletRequest;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Utility to manage application configuration values.
 */
public class ConfigUtil {
	
	private ConfigUtil () {}
	
	static private final Logger _LOG = Logger.getLogger("com.igzcode.java.web.util.ConfigUtil");
	
	static private boolean _IsLoaded = false;
	static private String _AppPath = null;
	
	static private HashMap<String, String> _Global;
	static private HashMap<String, String> _Local;
	static private HashMap<String, HashMap<String, String>> _Env;
	static private String _CurrentEnv;
	
	static private boolean _TestMode = false;
	
	static private String _BaseURL;
	
	static synchronized private void _LoadOnce () {
		if ( !_IsLoaded ) {
			_IsLoaded = true;
			try {
				_Load();
			}
			catch (Exception e) {
				_IsLoaded = false;
				e.printStackTrace();
				_LOG.severe("ConfigUtil can not load config.xml file.");
			}
		}
	}
	
	static private void _Load () throws ParserConfigurationException, SAXException, IOException, XPathExpressionException {
		
		// Read config.xml
		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
		//Document doc = docBuilder.parse( ClassLoader.getSystemResourceAsStream("config.xml") );
		Document doc = docBuilder.parse( ConfigUtil.class.getClassLoader().getResourceAsStream("config.xml") );
		
		// Prepare XPath
		XPathFactory factory = XPathFactory.newInstance();
		XPath xpath = factory.newXPath();
		
		// Get current environment
		XPathExpression expression = xpath.compile("/Configuration/Env/Current");
		_CurrentEnv = expression.evaluate(doc);

		// Load global values
		_Global = new HashMap<String, String>();
		
		expression = xpath.compile("/Configuration/Global/*");
		NodeList nodeList = (NodeList) expression.evaluate(doc, XPathConstants.NODESET);
		
		Node node;
		String label, value;
		for ( int i=0; i<nodeList.getLength(); i++ ) {
			node = nodeList.item(i);
			label = node.getNodeName();
			value = node.getAttributes().getNamedItem("value").getNodeValue();
			
			_Global.put(label, value);
		}

		// Load local values
		_Local = new HashMap<String, String>();
		
		expression = xpath.compile("/Configuration/Local/*");
		nodeList = (NodeList) expression.evaluate(doc, XPathConstants.NODESET);
		
		for ( int i=0; i<nodeList.getLength(); i++ ) {
			node = nodeList.item(i);
			label = node.getNodeName();
			value = node.getAttributes().getNamedItem(_CurrentEnv).getNodeValue();
			
			_Local.put(label, value);
		}
		
		// Load Env variables
		_Env = new HashMap<String, HashMap<String,String>>();
		
		expression = xpath.compile("/Configuration/Env/*");
		nodeList = (NodeList) expression.evaluate(doc, XPathConstants.NODESET);
		
		NamedNodeMap nodeMap;
		HashMap<String, String> attrMap;
		Node attr;
		for ( int i=0; i<nodeList.getLength(); i++ ) {
			node = nodeList.item(i);
			label = node.getNodeName();
			attrMap = new HashMap<String, String>();
			
			nodeMap = node.getAttributes();
			for ( int j=0; j<nodeMap.getLength(); j++ ) {
				attr = nodeMap.item(j);
				attrMap.put(attr.getNodeName(), attr.getNodeValue());
			}
			
			_Env.put(label, attrMap);
		}
		
	}
	
	/**
	 * Get a global configuration value.
	 * The global values are defined in config.xml file,
	 * a quick example:
	 * <pre>
	 * {@code
	 * <?xml version="1.0" encoding="UTF-8"?>
	 * <Configuration>
	 *     <Global>
	 *         <Example value="my value" />
	 *     </Global>
	 * </Configuration>
	 * }
	 * </pre>
	 * For example:
	 * <pre>
	 * {@code
	 * ConfigUtil.Global("Example"); // Return "my value"
	 * }
	 * </pre>
	 * @param p_label XML node name
	 * @return The configuration value
	 */
	static public String Global ( String p_label ) {
		_LoadOnce();
		
		return _Global.get(p_label);
	}
	
	/**
	 * Get a configuration value by the current environment.
	 * The current environment and the local values are defined
	 * in {@code config.xml} file, a quick example:
	 * <pre>
	 * {@code
	 * <?xml version="1.0" encoding="UTF-8"?>
	 * <Configuration>
	 *   <Env>
	 *      <Default>dev</Default>
	 *   </Env>
	 *   <Local>
	 *      <Example dev="dev value" pro="pro value" />
	 *   </Local>
	 * </Configuration>
	 * }
	 * </pre>
	 * 
	 * For example:
	 * <pre>
	 * {@code
	 * ConfigUtil.Local("Example"); // "dev value"
	 * }
	 * </pre>
	 * 
	 * @param p_label The XML node name
	 * @return The indicated configuration value
	 */
	static public String Local ( String p_label ) {
		_LoadOnce();
		
		return _Local.get(p_label);
	}
	
	/**
	 * Get a environment configuration value.
	 * These are defined in config.xml file, a quick example:
	 * <pre>
	 * {@code
	 * <?xml version="1.0" encoding="UTF-8"?>
	 * <Configuration>
	 *     <Env>
	 *         <DB usr="user" pwd="***" server="192.168.0.xx" dbname="database_name" dbtype="MYSQL" />
	 *     </Env>
	 * </Configuration>
	 * }
	 * </pre>
	 * 
	 * For example:
	 * <pre>
	 * {@code
	 * ConfigUtil.Env("DB", "dbtype"); // "MYSQL"
	 * }
	 * </pre>
	 * @param p_label The XML node name
	 * @param p_atribute The node attribute name
	 * @return The attribute value for the specificated environment configuration
	 */
	static public String Env ( String p_label, String p_attribute) {
		_LoadOnce();
		
		return _Env.get(p_label).get(p_attribute);
	}
	
	/**
	 * Return the current environment defined in the config.xml file.
	 * A quick example:
	 * <pre>
	 * {@code
	 * <?xml version="1.0" encoding="UTF-8"?>
	 * <Configuration>
	 *     <Env>
	 *         <Default>dev</Default>
	 *     </Env>
	 * </Configuration>
	 * }
	 * </pre>
	 * 
	 * For example:
	 * <pre>
	 * {@code ConfigUtil.Env(); // "dev"}
	 * </pre>
	 * @return The current environment code
	 */
	static public String CurrentEnv () {
		_LoadOnce();
		
		return _CurrentEnv;
	}
	
	/**
	 * Get the application file system path.
	 * @return The root application path
	 */
	static public String AppPath () {
		if ( _AppPath == null ) {
			_AppPath = System.getProperty("user.dir");
		}
		return _AppPath;
	}
	
	static public void TestMode() {
		_TestMode = true;
	}
	
	static public boolean IsTestModeEnabled() {
		return _TestMode;
	}

	public static String BaseURL(HttpServletRequest p_request) {
		
		if ( _BaseURL == null ) {
			if ( _TestMode ) {
				_BaseURL = "";
			}
			else {
				_BaseURL = p_request.getScheme()   + "://" +
						p_request.getServerName()  + ":"   + p_request.getServerPort() +
						p_request.getContextPath() + "/"
						;
			}
		}
		
		return _BaseURL;
	}

}
