package com.igzcode.java.web.util;

import java.io.IOException;
import java.util.HashMap;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.igzcode.java.util.StringUtil;
import com.igzcode.java.util.collection.NameValue;
import com.igzcode.java.util.collection.NameValueArray;

public class LangUtil {
	
	private LangUtil () {}
	
	static private final Logger _LOG = Logger.getLogger("com.igzcode.java.web.util.LangUtil");
	
	static private HashMap<String, HashMap<String, String>> _Lang = null;
	
	static synchronized private void _LoadOnce ( String p_locale ) {
		if ( _Lang == null ) {
			_Lang = new HashMap<String, HashMap<String,String>>();
		}
		
		if ( !_Lang.containsKey(p_locale) ) {
			try {
				_Load(p_locale);
			}
			catch (Exception e) {
				e.printStackTrace();
				_LOG.severe("LangUtil can not load \"" + p_locale + ".xml\" file.");
			}
		}
	}
	
	static private void _Load ( String p_locale ) throws ParserConfigurationException, SAXException, IOException, XPathExpressionException {
		
		// Read [locale].xml
		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
		Document doc = docBuilder.parse( ClassLoader.getSystemClassLoader().getResourceAsStream("lang/" + p_locale + ".xml") );

		// Prepare XPath
		XPathFactory factory = XPathFactory.newInstance();
		XPath xpath = factory.newXPath();
		
		// Load data
		HashMap<String, String> localeText = new HashMap<String, String>();
		XPathExpression expression = xpath.compile("/Lang/*");
		NodeList nodeList = (NodeList) expression.evaluate(doc, XPathConstants.NODESET);
		Node node;
		for ( int i=0; i<nodeList.getLength(); i++ ) {
			node = nodeList.item(i);
			
			if ( localeText.containsKey(node.getNodeName()) ) {
				_LOG.warning("LangUtil \"" + p_locale + ".xml\" has \"" + node.getNodeName() + "\" duplicated");
			}
			
			localeText.put(node.getNodeName(), node.getTextContent());
		}
		
		_Lang.put( p_locale,  localeText);
	}
	
	/**
	 * Get a translation text by locale and label.
	 * If the text not exists, return <b>{@code ???[p_locale]???}</b>.<br>
	 * The translations are defined in resource files {@code [locale].xml}
	 * with the following structure:
	 * <pre>
	 * {@code
	 * <Lang>
	 *    <Example>my example text</Example>
	 * <Lang>
	 * }
	 * </pre>
	 * Example:
	 * <pre>
	 * LangUtil.Text("en_US", "Example");
	 * // Return "my example text"
	 * }
	 * </pre>
	 * 
	 * @param p_locale Locale code like {@code es_ES}
	 * @param p_label Traduction label
	 * @return The translated text if exists
	 */
	static public String Text ( String p_locale, String p_label ) {
		return Text(p_locale, p_label, null);
	}
	
	/**
	 * Get a translation text by locale and label and replace the indicated params.<br>
	 * If the text not exists, return <b>{@code ???[p_locale]???}</b>.<br>
	 * The translations are defined in resource XML files {@code lang/[locale].xml}.<br>
	 * For example, {@code lang/en_US.xml}:
	 * <pre>
	 * {@code
	 * <?xml version="1.0" encoding="UTF-8"?>
	 * <Lang>
	 *    <Greetings>Hi @@NAME@@!</Greetings>
	 * <Lang>
	 * }
	 * </pre>
	 * 
	 * Example:
	 * <pre>
	 * NameValueArray params = new NameValueArray();
	 * params.Add("@@NAME@@", "John Doe");
	 * LangUtil.Text("en_US", "Greetings", params);
	 * // Return "Hi John Doe!"
	 * </pre>
	 * 
	 * @param p_locale Locale code like {@code es_ES}
	 * @param p_label Traduction label
	 * @param p_params A collection of parameters to replace
	 * @return Translated text if exists
	 */
	static public String Text ( String p_locale, String p_label, NameValueArray p_params ) {
		_LoadOnce(p_locale);
		
		String translation = "???" + p_label + "???";
		
		HashMap<String, String> translations = _Lang.get(p_locale);
		if ( translations != null ) {
			if ( !StringUtil.IsNullOrEmpty(translations.get(p_label)) ) {
				translation = translations.get(p_label);
			}
			
			if ( p_params != null && p_params.Size() > 0 ) {
				for ( NameValue nameValue : p_params ) {
					translation = translation.replaceAll(nameValue.GetName(), nameValue.GetValue().toString());
				}
			}
		}
		return translation;
	}

}
